#include  <exec/types.h>
#include  <stdlib.h>
#include  <string.h>
#include  <math.h>

#include  "PathFinderCalculs.h"


int Nb_Noeuds;        //nombre de noeuds passs en revue lors de la recherche
static char Map[ TAILLE_X ][ TAILLE_Y ];
StNoeud * Open,        //liste des noeuds  inspecter
       * Closed;      //liste des noeuds dj examins


//
// fonction recherchant Noeud dans Liste. Si trouv, retourne le noeud quivalant situ dans la liste
//
static StNoeud * RechercherNoeud( StNoeud * Noeud, StNoeud * Liste )
{
    StNoeud * Temp;        //permet de parcourir la liste

    for ( Temp = Liste; Temp != NULL; Temp = Temp->Suivant )
        if ( Temp->x == Noeud->x && Temp->y == Noeud->y )
            return( Temp );

    return( NULL );
}


//
// fonction calculant une estimation du cot entre deux points
// cette estimation ne doit pas tre survalue par rapport au cot du chemin optimal
// on choisit donc le cot correspondant  la distance euclidienne entre les deux points
//
static long PathCostEstimate( int x1, int y1, int x2, int y2 )
{
    return( ( long )( ( x2 - x1 ) ^ 2 + ( y2 - y1 ) ^ 2 ) );
}


//
// fonction calculant le cot entre deux noeuds ( donc entre deux cases de la map )
// ce cot peut dpendre du type de terrain ( terrain escarp, facile, ..., eau, marais, ... )
// mais peut aussi dpendre des amliorations souhaites ( adoucissement du chemin en alourdissant le cot si le chemin  une brisure )
//
static long CoutEntre2Points( StNoeud * N1, StNoeud * N2 )
{
    return( 1 );
}


//
// Initialisation des listes  NULL
//
void InitPathFinderCalculs( void )
{
    Open = NULL;
    Closed = NULL;
}


//
// initialisation de la map
//
void EffaceMap( void )
{
    memset( Map, CASE_VIDE, sizeof( Map ) );
}


//
// fonction retournant l'tat de la case souhaite
//
char GetCase( int x, int y )
{
    if ( x < 0 || x > TAILLE_X || y < 0 || y > TAILLE_Y )
        return( ERREUR );

    return( Map[ x ][ y ] );
}


//
// fonction modifiant l'tant de la case souhaite
//
void SetCase( int x, int y, char Valeur )
{
    if ( x >= 0 && x < TAILLE_X && y >= 0 && y < TAILLE_Y )
        Map[ x ][ y ] = Valeur;
}


//
// fonction examinant le voisin de Noeud situ  la position passe en paramtre
//
static void ExaminerVoisin( StNoeud * Noeud, int x, int y, int xGoal, int yGoal )
{
    long NewCost;         //cot depuis le dpart jusqu'au nouveau noeud
    StNoeud * NewNoeud,    //noeud  examiner
            * Temp;        //sert  la recherche d'un noeud dj existant

    if ( x < 0 || y < 0 || x >= TAILLE_X || y >= TAILLE_Y )
        return;    //on sort de la map -> retour en arrire

    if ( Map[ x ][ y ] == CASE_PLEINE )
        return;    //obstacle -> retour en arrire

    Nb_Noeuds++;
    NewNoeud = ( StNoeud * )malloc( sizeof( StNoeud ) );
    NewNoeud->x = x;
    NewNoeud->y = y;
    NewNoeud->Parent = Noeud;
    NewNoeud->Precedent = NewNoeud->Suivant = NULL;
    NewNoeud->CostFromStart = 0.0f;
    NewNoeud->CostToGoal = 0.0f;
    NewCost = Noeud->CostFromStart + CoutEntre2Points( Noeud, NewNoeud );

    //on regarde si le noeud  dj t rpertori ou si on a trouv un plus court chemin pour l'atteindre
    //recherche de NewNoeud dans Open
    Temp = RechercherNoeud( NewNoeud, Open );
    //si NewNoeud a dj t rpertori, on regarde si on a trouv un meilleur chemin
    if ( Temp )
        {
        if ( Temp->CostFromStart <= NewCost )
            {
            //aucune amlioration: on retourne en arrire
            free( NewNoeud );
            return;
            }
        else    //sinon supprime Temp
            {
            if ( Temp->Precedent == NULL )
                {
                Open = Temp->Suivant;
                Open->Precedent = NULL;
                }
            else
                {
                Temp->Precedent->Suivant = Temp->Suivant;
                if ( Temp->Suivant != NULL )
                    Temp->Suivant->Precedent = Temp->Precedent;
                }
            free( Temp );
            }
        }
    else
        {
        //recherche de NewNoeud dans Closed
        Temp = RechercherNoeud( NewNoeud, Closed );
        //si NewNoeud a dj t rpertori, on regarde si on a trouv un meilleur chemin
        if ( Temp != NULL )
            {
            if ( Temp->CostFromStart <= NewCost )
                {
                //aucune amlioration: on retourne en arrire
                free( NewNoeud );
                return;
                }
            else    //sinon supprime Temp
                {
                if ( Temp->Precedent == NULL )
                    {
                    Closed = Temp->Suivant;
                    Closed->Precedent = NULL;
                    }
                else
                    {
                    Temp->Precedent->Suivant = Temp->Suivant;
                    if ( Temp->Suivant != NULL )
                        Temp->Suivant->Precedent = Temp->Precedent;
                    }
                free( Temp );
                }
            }
        else
            {
            //NewNoeud est un nouveau noeud, ou on a trouv un meilleur chemin pour l'atteindre
            NewNoeud->CostFromStart = NewCost;
            NewNoeud->CostToGoal = PathCostEstimate( NewNoeud->x, NewNoeud->y, xGoal, yGoal );

            //on ajoute NewNoeud  la fin de Open
            if ( Open == NULL )
                Open = NewNoeud;
            else
                {
                if ( Open == NULL )
                    {
                    NewNoeud->Precedent = NULL;
                    NewNoeud->Suivant = NULL;
                    Open = NewNoeud;
                    }
                else
                    {
                    Temp = Open;
                    while ( Temp->Suivant != NULL )
                        Temp = Temp->Suivant;

                    Temp->Suivant = NewNoeud;
                    NewNoeud->Precedent = Temp;
                    NewNoeud->Suivant = NULL;
                    }
                }
            }
        }
}


//
// fonction calculant le chemin reliant Start  Goal en utilisant l'algorithme A*
// Diagonales: indique si les dplacements en diagonale sont autoriss
//
void AStarSearch( int xStart, int yStart, int xGoal, int yGoal, int Diagonales )
{
    StNoeud * Noeud,    //noeud  examiner
           * Temp;

    //cration du premier noeud dans Open
    Open = ( StNoeud * )malloc( sizeof( StNoeud ) );
    Open->x = xStart;
    Open->y = yStart;
    Open->Parent = NULL;
    Open->Precedent = NULL;
    Open->Suivant = NULL;
    Open->CostFromStart = 0.0f;
    Open->CostToGoal = PathCostEstimate( xStart, yStart, xGoal, yGoal );

    Nb_Noeuds = 0;

    //on lance l'algorithme A*
    while ( Open != NULL )
        {
        //on rcupre le noeud de plus faible poids  examiner
        Temp = Open;
        Noeud = Open;
        while ( Temp != NULL )
            {
            if ( Temp->CostFromStart<Noeud->CostFromStart )
                Noeud = Temp;

            Temp = Temp->Suivant;
            }

        //si le noeud est le point d'arrive, on a termin la recherche
        if ( Noeud->x == xGoal && Noeud->y == yGoal )
            return;
        else
            {
            //on dtache Noeud de Open
            if ( Noeud->Precedent == NULL )
                {
                Open = Noeud->Suivant;
                if ( Noeud->Suivant != NULL )
                    Open->Precedent = NULL;
                }
            else
                {
                Temp = Open;
                while( Temp != NULL )
                    {
                    if ( Temp->x == Noeud->x && Temp->y == Noeud->y )
                        {
                        Temp->Precedent->Suivant = Temp->Suivant;
                        if ( Temp->Suivant != NULL )
                            Temp->Suivant->Precedent = Temp->Precedent;

                        break;
                        }
                    Temp = Temp->Suivant;
                    }
                }
            //on examine chaque voisin de Noeud
            ExaminerVoisin( Noeud, Noeud->x-1, Noeud->y, xGoal, yGoal );    //voisin de gauche
            ExaminerVoisin( Noeud, Noeud->x+1, Noeud->y, xGoal, yGoal );    //voisin de droite
            ExaminerVoisin( Noeud, Noeud->x, Noeud->y-1, xGoal, yGoal );    //voisin du haut
            ExaminerVoisin( Noeud, Noeud->x, Noeud->y+1, xGoal, yGoal );    //voisin du bas
            //diagonales
            if ( Diagonales )
                {
                ExaminerVoisin( Noeud, Noeud->x-1, Noeud->y+1, xGoal, yGoal );    //voisin en bas  gauche
                ExaminerVoisin( Noeud, Noeud->x+1, Noeud->y+1, xGoal, yGoal );    //voisin en bas  droite
                ExaminerVoisin( Noeud, Noeud->x-1, Noeud->y-1, xGoal, yGoal );    //voisin en haut  gauche
                ExaminerVoisin( Noeud, Noeud->x+1, Noeud->y-1, xGoal, yGoal );    //voisin en haut  droite
                }
            }
        //on ajoute Noeud  la fin de Closed
        if ( Closed == NULL )
            Closed = Noeud;
        else
            {
            Temp = Closed;
            while ( Temp->Suivant != NULL )
                Temp = Temp->Suivant;

            Noeud->Precedent = Temp;
            Noeud->Suivant = NULL;
            Temp->Suivant = Noeud;
            }
        }
    //MessageBox( NULL, "Aucun chemin ne peut relier les deux points", "Erreur - PathFinding", MB_ICONINFORMATION );
}


//
//fonction librant la mmoire contenue dans les listes Open et Closed
//
void Liberer_Listes( void )
{
    StNoeud * Temp;

    Temp = Open;
    while ( Open != NULL )
        {
        Open = Open->Suivant;
        free( Temp );
        Temp = Open;
        }

    Temp = Closed;
    while ( Closed != NULL )
        {
        Closed = Closed->Suivant;
        free( Temp );
        Temp = Closed;
        }
}
